;;;; Liveness analysis
;;;; LastChange: Wed Apr 11 13:00:52 2007

;;; $B%G%b(B ESC x sck-live-demo RET
;;;
;;; * Input
;;; 
;;; <basic-blocks> ::= ((<Label> (<Inst>..) (<Succ>..))..)
;;; 
;;; <Label> ::= <String>
;;; <Inst> ::= (<Defs> <Refs>)
;;; <Defs> ::= (<Vreg>..)
;;; <Refs> ::= (<Vreg>..)
;;; <Vreg> ::= <Sexp>
;;; <Succ> ::= <Label>
;;; 
;;; 
;;; * Output
;;; 
;;; <liveness-info> ::= ((<Label> (<LiveIn>..) (<LiveOut>..))..)
;;;
;;; <LiveIn> ::= <Label>
;;; <LiveOut> ::= <Label>


(defconst sck=live-input-sample
  '((basic-blocks (("Lbeg"
		    ((((reg "i")) nil)
		     (((reg "m")) nil)
		     (((reg "n")) ((reg "m")))
		     (((reg "total")) nil)
		     (nil nil))
		    ("L1"))
		   ("L1"
		    ((((reg "sum")) ((reg "total") (reg "i")))
		     (((reg "total")) ((reg "sum")))
		     (((reg "i")) ((reg "i")))
		     (nil ((reg "i") (reg "n"))))
		    ("L1" "Lend"))
		   ("Lend"
		    ((nil ((reg "total"))))
		    nil)))))

(defun sck-live-demo ()
  (interactive)
  (let ((input sck=live-input-sample))
    (with-display-buffer "*sck-live*"
      (insertf "%s\n%s"
	       (sck-live-format-input input)
	       (sck-live-format-output (sck-live input))))))

(defun sck-live-format-input (input)
  (sck-letalist input
      (basic-blocks)
    (with-temp-buffer
      (insert "* Input:\n((basic-blocks  (")
      (dolist (b basic-blocks)
	(insertf "\n  (%S\n" (car b))
	(dolist (i (cadr b))
	  (insertf "    %S\n" i))
	(insertf "    %S)" (caddr b)))
      (insert ")))\n")
      (buffer-string))))

(defun sck-live-format-output (output)
  (sck-letalist output
      (liveness-info)
    (with-temp-buffer
      (insertf "* Output:\n((liveness-info  (")
      (dolist (b liveness-info)
	(insertf "\n  %S" b))
      (insertf ")))\n")
      (buffer-string))))
  

(defun sck-live (input)
  "Live analysis $B%H%C%W%l%Y%k!#(B"
  (sck-letalist input
      (basic-blocks)
    ;; Syntax check
    (sck-live-check-input input)
    ;; Live analysis
    (sck-live-bblocks-to-output
     (sck-live-analysis
      (sck-live-make-bblocks basic-blocks)))))

(defun sck-live-check-input (input)
  ;; $B!z(B check it and then return it.
  input)

(defstruct sck-live-bblock
  label ; jump $B%i%Y%k(B
  insts ; $B%i%Y%k0J2<$NL?Na$NNs(B
  succ  ; outgoing bblock $B$N%j%9%H(B
  pred  ; incomming bblock $B$N%j%9%H(B
  ;; $B0J2<$=$NB>$NMM!9$J>pJs(B
  ;; liveness
  livein
  liveout)

(defun sck-live-bblocks-to-output (bblocks)
  (list
   (list 'liveness-info
	 (mapcar #'(lambda (b)
		     (list (sck-live-bblock-label b)
			   (sck-live-bblock-livein b)
			   (sck-live-bblock-liveout b)))
		 bblocks))))

(defun sck-live-format-bblock (bblock-s)
  "bblock $B$^$?$O$=$N%j%9%H$r%U%)!<%^%C%H$9$k!#(B"
  (with-temp-buffer
    (cond
     ((sck-live-bblock-p bblock-s)
      (insertf "%S ; %S->%S ; livein=%s liveout=%s\n"
	       (sck-live-bblock-label bblock-s)
	       (mapcar 'sck-live-bblock-label
		       (sck-live-bblock-pred bblock-s))
	       (mapcar 'sck-live-bblock-label
		       (sck-live-bblock-succ bblock-s))
	       (sck-live-bblock-livein bblock-s)
	       (sck-live-bblock-liveout bblock-s))
      (dolist (inst (sck-live-bblock-insts bblock-s))
	(if (and (eq (car inst) 'setq)
		 (eq (cadr inst) (caddr inst)))
	    (insertf "  ;; %S\n" inst)
	    (insertf "  %S\n" inst))))
     (t
      (dolist (b bblock-s)
	(insertf "%s\n" (sck-live-format-bblock b)))))
    (buffer-string)))


(defun sck-live-make-bblocks (input-bbs)
  "input $B$N(B basic-blocks"
  (let ((bblocks (mapcar #'(lambda (b)
			     ;; b = (Label (inst...) (succ...))
			     (make-sck-live-bblock
			      :label (car b)
			      :insts (cadr b)))
			 input-bbs)))
    ;; succ, pred $B$N@_Dj!#(B
    (dolist (b bblocks)
      (setf (sck-live-bblock-succ b)
	    (mapcar #'(lambda (succlabel)
			(let ((bs (find succlabel bblocks
					:key 'sck-live-bblock-label
					:test 'equal)))
			  (push b (sck-live-bblock-pred bs))
			  bs))
		    (caddr (assoc (sck-live-bblock-label b) input-bbs)))))
    ;;
    bblocks))


(defun sck-live-analysis (bblocks)
  "bblock $B$N%j%9%H(B BBLOCKS $B$r<u$1<h$j!"=q$/(B bblock $B$N(B live{in,out} $B$r7W;;!#(B
BBLOCKS $B$rJV$9!#(B"
  (let ((first t)
	(changed t))
    ;; $B=i4|2=(B
    (dolist (b bblocks)
      (setf (sck-live-bblock-livein b) nil
	    (sck-live-bblock-liveout b) nil))
    ;; $B%U%m!<2r@O(B
    (while changed
      (setq changed nil)
      ;; bblock $B$r2<$+$i>e$X(B
      (dolist (b (reverse bblocks))
	;; (message "bblock %s" (sck-live-bblock-label b))
	(let ((liveout nil)
	      (changed1 nil))
	  ;; succ bblock $B$N(B livein $B$r(B b $B$N(B liveout $B$X(B $B"@!#(B
	  (setq liveout (sck-live-bblock-liveout b))
	  (dolist (bs (sck-live-bblock-succ b))
	    (setq liveout
		  (union liveout (sck-live-bblock-livein bs) :test 'equal)))
	  ;; $B"@(B $BA`:n$GA}$($?$i(B changed1 = t
	  (when (not (set-equal liveout (sck-live-bblock-liveout b)
				:test 'equal))
	    (setf (sck-live-bblock-liveout b) liveout)
	    (setq changed1 t))
	  ;; $B:G=i$N0l2sL\$+!"(Bchanged1 $B$J$iL?Na$r2<$+$i>e$X:F7W;;$7!"(Bchanged t$B!#(B
	  (when (or first changed1)
	    (dolist (inst (reverse (sck-live-bblock-insts b)))
	      (let* ((defs (car inst))
		     (refs (cadr inst)))
		;; def $B$rH4$-!"(Bref $B$rDI2C$9$k!#(B
		(setq liveout (set-difference liveout defs :test 'equal)
		      liveout (union liveout refs :test 'equal))))
	    (setf (sck-live-bblock-livein b) liveout)
	    (setq changed t))))
      ;; $B:G=i$N0l2s%U%i%0(B off$B!#(B
      (setq first nil))
    ;;
    bblocks))


;;;
(provide 'sck-live)
